/*
 *
 *	pg_migrator.c
 *
 */

#include "pg_migrator.h"

#ifdef HAVE_LANGINFO_H
#include <langinfo.h>
#endif


#define ALL_DUMP_FILE		"pg_migrator_dump_all.sql"
/* contains both global db information and CREATE DATABASE commands */
#define GLOBALS_DUMP_FILE	"pg_migrator_dump_globals.sql"
#define DB_DUMP_FILE		"pg_migrator_dump_db.sql"


static void install_support_functions(migratorContext *ctx);
static void uninstall_support_functions(migratorContext *ctx);
static void generate_old_dump(migratorContext *ctx);
static void split_old_dump(migratorContext *ctx);
static void copy_clog_xlog_xid(migratorContext *ctx);
static void set_frozenxids(migratorContext *ctx);
static void check_cluster_versions(migratorContext *ctx);
static void check_cluster_compatibility(migratorContext *ctx, bool live_check);
static void setup(migratorContext *ctx, char *argv0, bool live_check);
static void cleanup(migratorContext *ctx);
static void rename_old_pg_control(migratorContext *ctx);

/* OID system catalog preservation added during PG 8.5 development */
#define OID_PRESERVATION_CAT_VER	200912271

/*
 *	main
 */
int
main(int argc, char **argv)
{
	migratorContext ctx;
	char *sequence_script_file_name = NULL;
	bool	live_check = false;
	
	memset(&ctx, 0, sizeof(ctx));

	parseCommandLine(&ctx, argc, argv);

	if (ctx.check && is_server_running(&ctx, ctx.old.pgdata))
	{
		live_check = true;
		if (ctx.old.port == ctx.new.port)
			pg_log(&ctx, PG_FATAL, "When checking a live server, "
					"the old and new port numbers must be different.\n");
		pg_log(&ctx, PG_REPORT, "PerForming Consistency Checks on Old Live Server\n");
		pg_log(&ctx, PG_REPORT, "------------------------------------------------\n");
	}
	else
	{
		pg_log(&ctx, PG_REPORT, "Performing Consistency Checks\n");
		pg_log(&ctx, PG_REPORT, "-----------------------------\n");
	}

	setup(&ctx, argv[0], live_check);

	check_cluster_versions(&ctx);
	check_cluster_compatibility(&ctx, live_check);


	/* -- OLD -- */

	if (!live_check)
		start_postmaster(&ctx, CLUSTER_OLD, false);

	set_locale_and_encoding(&ctx, CLUSTER_OLD);
	
	get_pg_database_relfilenode(&ctx, CLUSTER_OLD);

	/* Extract a list of databases and tables from the old cluster */
	get_db_and_rel_infos(&ctx, &ctx.old.dbarr, CLUSTER_OLD);

	get_tablespace_paths(&ctx);
	get_loadable_libraries(&ctx);

	/*
	 *	Check for various failure cases
	 */

	if (ctx.new.controldata.cat_ver < OID_PRESERVATION_CAT_VER)
	{
		v_8_4_check_for_composite_types(&ctx, CLUSTER_OLD);
		v_8_4_check_for_array_types(&ctx, CLUSTER_OLD);
		v_8_4_check_for_enum_types(&ctx, CLUSTER_OLD);
	}

	v8_3_check_for_isn_and_int8_passing_mismatch(&ctx, CLUSTER_OLD);
	
	if (GET_MAJOR_VERSION(ctx.old.pg_version) <= 803 &&
		GET_MAJOR_VERSION(ctx.new.pg_version) >= 804)
	{
		v8_3_check_for_name_data_type_usage(&ctx, CLUSTER_OLD);
		v8_3_check_for_tsquery_usage(&ctx, CLUSTER_OLD);
		if (ctx.check)
		{
			v8_3_rebuild_tsvector_tables(&ctx, true, CLUSTER_OLD);
			v8_3_invalidate_hash_gin_indexes(&ctx, true, CLUSTER_OLD);
			v8_3_invalidate_bpchar_pattern_ops_indexes(&ctx, true, CLUSTER_OLD);
		}
		else
			/*
			 *	While we have the old server running, create the script
			 *	to properly restore its sequence values but we report this
			 *	at the end.
			 */
			sequence_script_file_name =
				v8_3_create_sequence_script(&ctx, CLUSTER_OLD);
	}
	if (GET_MAJOR_VERSION(ctx.old.pg_version) <= 804 &&
		GET_MAJOR_VERSION(ctx.new.pg_version) >= 805)
		v8_4_populate_pg_largeobject_metadata(&ctx, true, CLUSTER_OLD);

	/* While not a check option, we do this while the old server is running */
	if (!ctx.check)
	{
		generate_old_dump(&ctx);
		split_old_dump(&ctx);
	}

	if (!live_check)
		stop_postmaster(&ctx, false, false);


	/* -- NEW -- */
	start_postmaster(&ctx, CLUSTER_NEW, false);

	set_locale_and_encoding(&ctx, CLUSTER_NEW);
	
	check_new_db_is_empty(&ctx);

	check_loadable_libraries(&ctx);

	check_locale_and_encoding(&ctx, &ctx.old.controldata, &ctx.new.controldata);

	if (ctx.transfer_mode == TRANSFER_MODE_LINK)
		check_hard_link(&ctx);
	
	if (ctx.check)
	{
		pg_log(&ctx, PG_REPORT, "\n*Clusters are compatible*\n");
		/* stops new cluster */
		stop_postmaster(&ctx, false, false);
		exit_nicely(&ctx, false);
	}

	pg_log(&ctx, PG_REPORT, "\n"
		"| If pg_migrator fails after this point, you must\n"
		"| re-initdb the new cluster before continuing.\n"
		"| You will also need to remove the \".old\" suffix\n"
		"| from old $PGDATA/global/pg_control");
	if (ctx.num_tablespaces)
		pg_log(&ctx, PG_REPORT,
			" and old\n| tablespace directory names");
	pg_log(&ctx, PG_REPORT, ".\n");

	pg_log(&ctx, PG_REPORT, "\nPerforming Migration\n");
	pg_log(&ctx, PG_REPORT, "--------------------\n");
	
	/*
	 *	Destructive Changes to Old Cluster (New cluster still running)
	 */

	/* rename pg_control so old server cannot be accidentally started */
	rename_old_pg_control(&ctx);
	
	/*
	 *	Rename all tablespace paths to .old so we can copy them later.
	 */
	rename_tablespaces(&ctx);
	check_ok(&ctx);

	/*
	 *	Non-Destructive Changes to New Cluster
	 */

	/*
	 * It would make more sense to freeze after loading the schema, but that
	 * would cause us to lose the frozenids restored by the load.
	 * We use --analyze so autovacuum doesn't update statistics later
	 */
	prep_status(&ctx, "Analyzing all rows in the new cluster");
	exec_prog(&ctx, true,
			  SYSTEMQUOTE "\"%s/vacuumdb\" --port %d --all --analyze >> %s 2>&1" SYSTEMQUOTE,
			  ctx.new.bindir, ctx.new.port, ctx.logfile);
	check_ok(&ctx);

	/*
	 * We do freeze after analyze so pg_statistic is also frozen
	 */
	prep_status(&ctx, "Freezing all rows on the new cluster");
	exec_prog(&ctx, true,
			  SYSTEMQUOTE "\"%s/vacuumdb\" --port %d --all --freeze >> %s 2>&1" SYSTEMQUOTE,
			  ctx.new.bindir, ctx.new.port, ctx.logfile);
	check_ok(&ctx);

	get_pg_database_relfilenode(&ctx, CLUSTER_NEW);

	stop_postmaster(&ctx, false, false);

	/*
	 *	Destructive Changes to New Cluster
	 */

	copy_clog_xlog_xid(&ctx);
	/* New now using xid of old system */


	/* -- NEW -- */
	start_postmaster(&ctx, CLUSTER_NEW, false);

	/*
	 *	We set autovacuum_freeze_max_age to its maximum value so autovacuum does
	 *	not launch here and delete clog files, before the frozen xids are set.
	 */

	set_frozenxids(&ctx);

	/*
	 * We have to create the databases first so we can create the
	 * toast table placeholder relfiles.
	 */
	prep_status(&ctx, "Creating databases in the new cluster");
	exec_prog(&ctx, true,
			  SYSTEMQUOTE "\"%s/%s\" --set ON_ERROR_STOP=on --port %d "
			  "-f \"%s/%s\" --dbname template1 >> \"%s\"" SYSTEMQUOTE,
			  ctx.new.bindir, ctx.new.psql_exe, ctx.new.port,
			  ctx.home_dir, GLOBALS_DUMP_FILE, ctx.logfile);
	check_ok(&ctx);

	get_db_and_rel_infos(&ctx, &ctx.new.dbarr, CLUSTER_NEW);

	stop_postmaster(&ctx, false, false);

	prep_status(&ctx, "Creating placeholder relfiles for toast relations\n");
	reserve_relfilenode_names(&ctx);
	prep_status(&ctx, "");
	check_ok(&ctx);


	/* -- NEW -- */
	start_postmaster(&ctx, CLUSTER_NEW, false);

	install_support_functions(&ctx);

	prep_status(&ctx, "Restoring database schema to new cluster");
	exec_prog(&ctx, true,
			  SYSTEMQUOTE "\"%s/%s\" --set ON_ERROR_STOP=on --port %d "
			  "-f \"%s/%s\" --dbname template1 >> \"%s\"" SYSTEMQUOTE,
			  ctx.new.bindir, ctx.new.psql_exe, ctx.new.port,
			  ctx.home_dir, DB_DUMP_FILE, ctx.logfile);
	check_ok(&ctx);

	/* regenerate now that we have db schemas */
	dbarr_free(&ctx.new.dbarr);
	get_db_and_rel_infos(&ctx, &ctx.new.dbarr, CLUSTER_NEW);

	prep_status(&ctx, "Restoring relations to use fixed toast file names\n");
	adjust_rels_to_use_fixed_toast_relfilenodes(&ctx, &ctx.old.dbarr, &ctx.new.dbarr);
	prep_status(&ctx, "");	/* in case nothing printed */
	check_ok(&ctx);

	uninstall_support_functions(&ctx);

	/*
	 * adjust_rels_to_use_fixed_toast_relfilenodes would have created new
	 * entries in the new database; hence regenerate new_dbinfos to get in sync.
	 */
	dbarr_free(&ctx.new.dbarr);
	get_db_and_rel_infos(&ctx, &ctx.new.dbarr, CLUSTER_NEW);

	stop_postmaster(&ctx, false, false);


	prep_status(&ctx, "Restoring user relation files\n");
	transfer_all_new_dbs(&ctx, &ctx.old.dbarr, &ctx.new.dbarr,
						 ctx.old.pgdata, ctx.new.pgdata);
	prep_status(&ctx, "");	/* in case nothing printed */
	check_ok(&ctx);

	/*
	 * Assuming OIDs are only used in system tables, there is no need to
	 * restore the OID counter because we have not transferred any OIDs from
	 * the old system, but we do it anyway just in case.  We do it late here
	 * because there is no need to have the schema load use new oids.
	 */
	prep_status(&ctx, "Setting next oid for new cluster");
	exec_prog(&ctx, true, SYSTEMQUOTE "\"%s/pg_resetxlog\" -o %u \"%s\" > " DEVNULL SYSTEMQUOTE,
		  ctx.new.bindir, ctx.old.controldata.chkpnt_nxtoid, ctx.new.pgdata);
	check_ok(&ctx);

	if (GET_MAJOR_VERSION(ctx.old.pg_version) <= 803 &&
		GET_MAJOR_VERSION(ctx.new.pg_version) >= 804)
	{
		start_postmaster(&ctx, CLUSTER_NEW, true);

		/* restore proper sequence values using file created from old server */
		if (sequence_script_file_name)
		{
			prep_status(&ctx, "Adjusting sequences");
			exec_prog(&ctx, true,
					  SYSTEMQUOTE "\"%s/%s\" --set ON_ERROR_STOP=on --port %d "
					  "-f \"%s\" --dbname template1 >> \"%s\"" SYSTEMQUOTE,
					  ctx.new.bindir, ctx.new.psql_exe, ctx.new.port,
					  sequence_script_file_name, ctx.logfile);
			unlink(sequence_script_file_name);
			pg_free(sequence_script_file_name);
			check_ok(&ctx);
		}
	
		v8_3_rebuild_tsvector_tables(&ctx, false, CLUSTER_NEW);
		v8_3_invalidate_hash_gin_indexes(&ctx, false, CLUSTER_NEW);
		v8_3_invalidate_bpchar_pattern_ops_indexes(&ctx, false, CLUSTER_NEW);
		stop_postmaster(&ctx, false, true);
	}

	if (GET_MAJOR_VERSION(ctx.old.pg_version) <= 804 &&
		GET_MAJOR_VERSION(ctx.new.pg_version) >= 805)
	{
		start_postmaster(&ctx, CLUSTER_NEW, true);
		v8_4_populate_pg_largeobject_metadata(&ctx, false, CLUSTER_NEW);
		stop_postmaster(&ctx, false, true);
	}
	
	pg_log(&ctx, PG_REPORT, "*Upgrade complete*\n");

	if (GET_MAJOR_VERSION(ctx.old.pg_version) >= 804 &&
		GET_MAJOR_VERSION(ctx.new.pg_version) >= 804)
		pg_log(&ctx, PG_REPORT, "\n"
			   "| Optimizer statistics is not transferred by pg_migrator\n"
			   "| so consider running ANALYZE on all databases in the\n"
			   "| newly-upgraded cluster.\n");
	else
		pg_log(&ctx, PG_REPORT, "\n"
		  		"| Optimizer statistics and free space information\n"
				"| are not transferred by pg_migrator so consider\n"
				"| running:\n"
				"| \tvacuumdb --all --analyze\n" 
				"| on the newly-upgraded cluster.\n");

	cleanup(&ctx);

	return 0;
}


/*
 *	generate_old_dump
 */
static void
generate_old_dump(migratorContext *ctx)
{
	/* run new pg_dumpall binary */
	prep_status(ctx, "Creating catalog dump");

	/*
	 * --binary-upgrade records the width of dropped columns in pg_class, and
	 * restores the frozenid's for databases and relations.
	 */
	exec_prog(ctx, true,
			SYSTEMQUOTE "\"%s/pg_dumpall\" --port %d --schema-only "
			"--binary-upgrade > \"%s/" ALL_DUMP_FILE "\"" SYSTEMQUOTE,
		    ctx->new.bindir, ctx->old.port, ctx->home_dir);
	check_ok(ctx);
}



/*
 *	split_old_dump
 *
 *	This function splits pg_dumpall output into global values and
 *	database creation, and per-db schemas.  This allows us to create
 *	the toast place holders between restoring these two parts of the
 *	dump.  We split on the first "\connect " after a CREATE ROLE
 *	username match;  this is where the per-db restore starts.
 *	
 *	We suppress recreation of our own username so we don't generate
 *	an error during restore
 */
static void
split_old_dump(migratorContext *ctx)
{
	FILE	*all_dump, *globals_dump, *db_dump;
	FILE	*current_output;
	char	line[LINE_ALLOC];
	bool	start_of_line = true;
	char	create_role_str[MAX_STRING];
	char	create_role_str_quote[MAX_STRING];
	char	filename[MAX_STRING];
	bool	suppressed_username = false;
	
	snprintf(filename, sizeof(filename), "%s/%s", ctx->home_dir, ALL_DUMP_FILE);
	if ((all_dump =	fopen(filename, "r")) == NULL)
		pg_log(ctx, PG_FATAL, "Cannot open dump file %s\n", filename);
	snprintf(filename, sizeof(filename), "%s/%s", ctx->home_dir, GLOBALS_DUMP_FILE);
	if ((globals_dump =	fopen(filename, "w")) == NULL)
		pg_log(ctx, PG_FATAL, "Cannot write to dump file %s\n", filename);
	snprintf(filename, sizeof(filename), "%s/%s", ctx->home_dir, DB_DUMP_FILE);
	if ((db_dump =	fopen(filename, "w")) == NULL)
		pg_log(ctx, PG_FATAL, "Cannot write to dump file %s\n", filename);
	current_output = globals_dump;
	
	/* patterns used to prevent our own username from being recreated */
	snprintf(create_role_str, sizeof(create_role_str),
			 "CREATE ROLE %s;", ctx->user);
	snprintf(create_role_str_quote, sizeof(create_role_str_quote),
			 "CREATE ROLE %s;", quote_identifier(ctx, ctx->user));
		
	while(fgets(line, sizeof(line), all_dump) != NULL)
	{
		/* switch to db_dump file output? */
		if (current_output == globals_dump && start_of_line &&
			suppressed_username &&
			strncmp(line, "\\connect ", strlen("\\connect ")) == 0)
			current_output = db_dump;

		/* output unless we are recreating our own username */
		if (current_output != globals_dump || !start_of_line ||
			(strncmp(line, create_role_str, strlen(create_role_str)) != 0 &&
			 strncmp(line, create_role_str_quote, strlen(create_role_str_quote)) != 0))
				fputs(line, current_output);
		else
			suppressed_username = true;
					
		if (strlen(line) > 0 && line[strlen(line) - 1] == '\n')
			start_of_line = true;
		else
			start_of_line = false;
	}	

	fclose(all_dump);
	fclose(globals_dump);
	fclose(db_dump);
}



/*
 *	copy_clog_xlog_xid()
 */
static void
copy_clog_xlog_xid(migratorContext *ctx)
{
	char		old_clog_path[MAXPGPATH];
	char		new_clog_path[MAXPGPATH];

	/* copy old commit logs to new data dir */
	prep_status(ctx, "Deleting new commit clogs");

	snprintf(old_clog_path, sizeof(old_clog_path), "%s/pg_clog", ctx->old.pgdata);
	snprintf(new_clog_path, sizeof(new_clog_path), "%s/pg_clog", ctx->new.pgdata);
	if (rmtree(new_clog_path, true) != true)
		pg_log(ctx, PG_FATAL, "Unable to delete directory %s\n", new_clog_path);
	check_ok(ctx);

	prep_status(ctx, "Copying old commit clogs to new server");
	/* libpgport's copydir() doesn't work in FRONTEND code */
#ifndef WIN32
	exec_prog(ctx, true, SYSTEMQUOTE "%s \"%s\" \"%s\"" SYSTEMQUOTE,
				"cp -Rf",
#else
				/* flags: everything, no confirm, quiet, overwrite read-only */
	exec_prog(ctx, true, SYSTEMQUOTE "%s \"%s\" \"%s\\\"" SYSTEMQUOTE,
				"xcopy /e /y /q /r",
#endif
				 old_clog_path, new_clog_path);
	check_ok(ctx);

	/* set the next transaction id of the new cluster */
	prep_status(ctx, "Setting next transaction id for new cluster");
	exec_prog(ctx, true, SYSTEMQUOTE "\"%s/pg_resetxlog\" -f -x %u \"%s\" > " DEVNULL SYSTEMQUOTE,
	   ctx->new.bindir, ctx->old.controldata.chkpnt_nxtxid, ctx->new.pgdata);
	check_ok(ctx);

	/* now reset the wal archives in the new cluster */
	prep_status(ctx, "Resetting WAL archives");
	exec_prog(ctx, true, SYSTEMQUOTE "\"%s/pg_resetxlog\" -l %u,%u,%u \"%s\" >> \"%s\" 2>&1" SYSTEMQUOTE,
			  ctx->new.bindir, ctx->old.controldata.chkpnt_tli,
			  ctx->old.controldata.logid, ctx->old.controldata.nxtlogseg,
			  ctx->new.pgdata, ctx->logfile);
	check_ok(ctx);
}


/*
 *	set_frozenxids()
 *
 *	We have frozen all xids, so set relfrozenxid and datfrozenxid
 *	to be the old cluster's xid counter, which we just set in the new
 *	cluster.  User-table frozenxid values will be set by pg_dumpall
 *	--binary-upgrade, but objects not set by the pg_dump must have
 *	proper frozen counters.
 */
static void	set_frozenxids(migratorContext *ctx)
{
	int			dbnum;
	PGconn	   *conn;
	PGresult   *dbres;
	int			ntups;
	
	prep_status(ctx, "Setting frozenxid counters in new cluster");

	conn = connectToServer(ctx, "template1", CLUSTER_NEW);

	/* set pg_database.datfrozenxid */
	PQclear(executeQueryOrDie(ctx, conn,
							"UPDATE pg_catalog.pg_database "
							"SET	datfrozenxid = '%u' "
							/* cannot connect to 'template0', so ignore */
							"WHERE	datname != 'template0'",
							ctx->old.controldata.chkpnt_nxtxid));

	/* get database names */
	dbres = executeQueryOrDie(ctx, conn,
							"SELECT	datname "
							"FROM	pg_catalog.pg_database "
							"WHERE	datname != 'template0'");

	/* free dbres below */
	PQfinish(conn);
	
	ntups = PQntuples(dbres);
	for (dbnum = 0; dbnum < ntups; dbnum++)
	{
		conn = connectToServer(ctx,  PQgetvalue(dbres, dbnum, 0), CLUSTER_NEW);

		/* set pg_class.relfrozenxid */
		PQclear(executeQueryOrDie(ctx, conn,
								"UPDATE	pg_catalog.pg_class "
								"SET	relfrozenxid = '%u' "
								/* only heap and TOAST are vacuumed */
								"WHERE	relkind = 'r' OR "
								"		relkind = 't'",
								ctx->old.controldata.chkpnt_nxtxid));
		PQfinish(conn);
	}

	PQclear(dbres);

	check_ok(ctx);
}


/*
 *	setup()
 */
static void
setup(migratorContext *ctx, char *argv0, bool live_check)
{
	char		exec_path[MAXPGPATH];	/* full path to my executable */

	/* disallow execution by the root user	*/
	pg_free(get_user_id(ctx));

	/*
	 * make sure the user has a clean environment, otherwise, we may confuse
	 * libpq when we connect to one (or both) of the servers.
	 */
	check_for_libpq_envvars(ctx);

	verify_directories(ctx);

	/* no postmasters should be running */
	if (!live_check && is_server_running(ctx, ctx->old.pgdata))
	{
		pg_log(ctx, PG_FATAL, "There seems to be a postmaster servicing the old cluster.\n"
					  "Please shutdown that postmaster and try again.\n");
	}

	/* same goes for the new postmaster */
	if (is_server_running(ctx, ctx->new.pgdata))
	{
		pg_log(ctx, PG_FATAL, "There seems to be a postmaster servicing the new cluster.\n"
			   "Please shutdown that postmaster and try again.\n");
	}

	/* get path to pg_migrator executable */
	if (find_my_exec(argv0, exec_path) < 0)
		pg_log(ctx, PG_FATAL, "Could not get pathname to pg_migrator: %s\n", strerror(errno));

	/* Trim off program name and keep just path */
	*last_dir_separator(exec_path) = '\0';
	canonicalize_path(exec_path);
	ctx->exec_path = pg_strdup(ctx, exec_path);
}


/*
 *	check_cluster_versions()
 */
static void
check_cluster_versions(migratorContext *ctx)
{
	/* get old and new cluster versions */
	ctx->old.pg_version = get_postgres_version(ctx, &ctx->old.pg_version_str, CLUSTER_OLD);
	ctx->new.pg_version = get_postgres_version(ctx, &ctx->new.pg_version_str, CLUSTER_NEW);

	/* We allow migration from/to the same major version for beta upgrades */

	if (GET_MAJOR_VERSION(ctx->old.pg_version) < 803)
		pg_log(ctx, PG_FATAL, "This utility can only upgrade from PostgreSQL version 8.3 and later.\n");

	if (!ctx->new.is_edb_as)
	{
		/* Only PG 8.4+ has enhancements to pg_dump, pg_dumpall, and vacuumdb */
		if (GET_MAJOR_VERSION(ctx->new.pg_version) < 804)
			pg_log(ctx, PG_FATAL, "This utility can only upgrade to PostgreSQL version 8.4 and later.\n");
	}
	else
	{
		if (GET_MAJOR_VERSION(ctx->new.pg_version) < 803)
			pg_log(ctx, PG_FATAL, "This utility can only upgrade to EnterpriseDB Postgres Plus Advanced Server\nversion 8.3 and later.\n");
	}

	/*
	 * We can't allow downgrading because we use the target pg_dumpall, and
	 * pg_dumpall cannot operate on new datbase versions, only older versions.
	 */
	if (ctx->old.pg_version > ctx->new.pg_version)
		pg_log(ctx, PG_FATAL, "This utility cannot be used to downgrade to older major PostgreSQL versions.\n");

	/* Must be built using _new_ server source */
	if (GET_MAJOR_VERSION(PG_VERSION_NUM) !=
		GET_MAJOR_VERSION(ctx->new.pg_version))
		pg_log(ctx, PG_FATAL,
				"This binary was built using %s PostgreSQL sources; it must be\n"
				"built using %s PostgreSQL sources to match the new cluster.\n",
#if defined(PG_MAJORVERSION) || !defined(EDB_NATIVE_LANG)
				PG_MAJORVERSION,
#else
				"8.3",	/* assume EDB AS 8.3 */
#endif
				ctx->new.pg_version_str);
}


/*
 *	check_cluster_compatibility()
 */
static void
check_cluster_compatibility(migratorContext *ctx, bool live_check)
{
	char		libfile[MAXPGPATH];
	FILE 		*lib_test;

	/*
	 *	Test pg_migrator.so is in the proper place.   We cannot copy it
	 *	ourselves because install directories are typically root-owned.
	 */
	snprintf(libfile, sizeof(libfile), "%s/pg_migrator%s", ctx->new.libpath,
				DLSUFFIX);

	if ((lib_test = fopen(libfile, "r")) == NULL)
		pg_log(ctx, PG_FATAL,
			"\npg_migrator%s must be created and installed in %s\n", DLSUFFIX, libfile);
	else
		fclose(lib_test);

	/* get/check pg_control data of servers */
	get_control_data(ctx, &ctx->old, live_check);
	get_control_data(ctx, &ctx->new, false);
	check_control_data(ctx, &ctx->old.controldata, &ctx->new.controldata);
}


/*
 *	cleanup()
 */
static void
cleanup(migratorContext *ctx)
{
	int			tblnum;
	char	filename[MAX_STRING];

	for (tblnum = 0; tblnum < ctx->num_tablespaces; tblnum++)
		pg_free(ctx->tablespaces[tblnum]);
	pg_free(ctx->tablespaces);

	dbarr_free(&ctx->old.dbarr);
	dbarr_free(&ctx->new.dbarr);
	pg_free(ctx->logfile);
	pg_free(ctx->user);
	pg_free(ctx->old.pg_version_str);
	pg_free(ctx->new.pg_version_str);
	pg_free(ctx->old.controldata.lc_collate);
	pg_free(ctx->new.controldata.lc_collate);
	pg_free(ctx->old.controldata.lc_ctype);
	pg_free(ctx->new.controldata.lc_ctype);
	pg_free(ctx->old.controldata.encoding);
	pg_free(ctx->new.controldata.encoding);
	
	if (ctx->log_fd != NULL)
	{
		fclose(ctx->log_fd);
		ctx->log_fd = NULL;
	}

	if (ctx->debug_fd)
		fclose(ctx->debug_fd);

	snprintf(filename, sizeof(filename), "%s/%s", ctx->home_dir, ALL_DUMP_FILE);
	unlink(filename);
	snprintf(filename, sizeof(filename), "%s/%s", ctx->home_dir, GLOBALS_DUMP_FILE);
	unlink(filename);
	snprintf(filename, sizeof(filename), "%s/%s", ctx->home_dir, DB_DUMP_FILE);
	unlink(filename);
}


/*
 * install_support_functions()
 *
 * pg_migrator requires some support functions, which should be dynamically
 * linked with the server. This routine installs those functions.
 */
static void
install_support_functions(migratorContext *ctx)
{
	int			dbnum;

	prep_status(ctx, "Adding support functions to new cluster");

	for (dbnum = 0; dbnum < ctx->new.dbarr.ndbs; dbnum++)
	{
		DbInfo	   *newdb = &ctx->new.dbarr.dbs[dbnum];
		PGconn	   *conn = connectToServer(ctx, newdb->db_name, CLUSTER_NEW);

		/* suppress NOTICE of dropped objects */
		PQclear(executeQueryOrDie(ctx, conn,
					"SET client_min_messages = warning;"));
		PQclear(executeQueryOrDie(ctx, conn,
					"DROP SCHEMA IF EXISTS binary_upgrade CASCADE;"));
		PQclear(executeQueryOrDie(ctx, conn,
					"RESET client_min_messages;"));

		PQclear(executeQueryOrDie(ctx, conn,
					"CREATE SCHEMA binary_upgrade;"));

		PQclear(executeQueryOrDie(ctx, conn,
					"CREATE OR REPLACE FUNCTION binary_upgrade.pg_toasttbl_recreate(OID,OID) "
					"RETURNS VOID "
					"AS '$libdir/pg_migrator' "
					"LANGUAGE C STRICT;"));
		PQclear(executeQueryOrDie(ctx, conn,
					"CREATE OR REPLACE FUNCTION binary_upgrade.pg_toasttbl_drop(OID) "
					"RETURNS VOID "
					"AS '$libdir/pg_migrator' "
					"LANGUAGE C STRICT;"));
		if (ctx->new.controldata.cat_ver >= OID_PRESERVATION_CAT_VER)
		{
			PQclear(executeQueryOrDie(ctx, conn,
						"CREATE OR REPLACE FUNCTION "
						"		binary_upgrade.set_next_pg_type_oid(OID) "
						"RETURNS VOID "
						"AS '$libdir/pg_migrator_sysoids' "
						"LANGUAGE C STRICT;"));
			PQclear(executeQueryOrDie(ctx, conn,
						"CREATE OR REPLACE FUNCTION "
						"		binary_upgrade.set_next_pg_type_array_oid(OID) "
						"RETURNS VOID "
						"AS '$libdir/pg_migrator_sysoids' "
						"LANGUAGE C STRICT;"));
			PQclear(executeQueryOrDie(ctx, conn,
						"CREATE OR REPLACE FUNCTION "
						"		binary_upgrade.set_next_pg_type_toast_oid(OID) "
						"RETURNS VOID "
						"AS '$libdir/pg_migrator_sysoids' "
						"LANGUAGE C STRICT;"));
			PQclear(executeQueryOrDie(ctx, conn,
						"CREATE OR REPLACE FUNCTION "
						"		binary_upgrade.add_pg_enum_label(OID, OID, NAME) "
						"RETURNS VOID "
						"AS '$libdir/pg_migrator_sysoids' "
						"LANGUAGE C STRICT;"));
		}
		PQfinish(conn);
	}
	check_ok(ctx);
}



/*
 * uninstall_support_functions()
 *
 * remove pg_migrator support functions
 */
static void
uninstall_support_functions(migratorContext *ctx)
{
	int			dbnum;

	prep_status(ctx, "Removing support functions from new cluster");

	for (dbnum = 0; dbnum < ctx->new.dbarr.ndbs; dbnum++)
	{
		DbInfo	   *newdb = &ctx->new.dbarr.dbs[dbnum];
		PGconn	   *conn = connectToServer(ctx, newdb->db_name, CLUSTER_NEW);

		/* suppress NOTICE of dropped objects */
		PQclear(executeQueryOrDie(ctx, conn,
					"SET client_min_messages = warning;"));
		PQclear(executeQueryOrDie(ctx, conn,
					"DROP SCHEMA binary_upgrade CASCADE;"));
		PQclear(executeQueryOrDie(ctx, conn,
					"RESET client_min_messages;"));
		PQfinish(conn);
	}
	check_ok(ctx);
}



static void
rename_old_pg_control(migratorContext *ctx)
{
	char	old_path[MAX_STRING], new_path[MAX_STRING];

	prep_status(ctx, "Adding \".old\" suffix to old global/pg_control");

	snprintf(old_path, sizeof(old_path), "%s/global/pg_control", ctx->old.pgdata);
	snprintf(new_path, sizeof(new_path), "%s/global/pg_control.old", ctx->old.pgdata);
	if (pg_mv_file(old_path, new_path) != 0)
	{
#ifdef WIN32
		_dosmaperr(GetLastError());
#endif
		pg_log(ctx, PG_FATAL, "Unable to rename %s to %s.\n", old_path, new_path);
	}
	check_ok(ctx);
}
